// WL_DRAW.C

#include "wl_def.h"
#pragma hdrstop

#include "wl_cloudsky.h"
#include "wl_atmos.h"
#include "wl_shade.h"

/*
=============================================================================

                               LOCAL CONSTANTS

=============================================================================
*/

// the door is the last picture before the sprites
#define DOORWALL        (PMSpriteStart-20)

#define ACTORSIZE       0x4000

/*
=============================================================================

                              GLOBAL VARIABLES

=============================================================================
*/

static byte *vbuf = NULL;
unsigned vbufPitch = 0;

int32_t    lasttimecount;
int32_t    frameon;
int         messagetime = 0;  
boolean fpscounter;

int fps_frames=0, fps_time=0, fps=0;

int *wallheight;
int min_wallheight;

//
// math tables
//
short *pixelangle;
int32_t finetangent[FINEANGLES/4];
fixed sintable[ANGLES+ANGLES/4];
fixed *costable = sintable+(ANGLES/4);

//
// refresh variables
//
fixed   viewx,viewy;                    // the focal point
short   viewangle;
fixed   viewsin,viewcos;

void    TransformActor (objtype *ob);
void    BuildTables (void);
void    ClearScreen (void);
int     CalcRotate (objtype *ob);
void    DrawScaleds (void);
void    CalcTics (void);
void    ThreeDRefresh (void);



//
// wall optimization variables
//
int     lastside;               // true for vertical
int32_t    lastintercept;
int     lasttilehit;
int     lasttexture;

//
// ray tracing variables
//
short    focaltx,focalty,viewtx,viewty;
longword xpartialup,xpartialdown,ypartialup,ypartialdown;

short   midangle,angle;

word    tilehit;
int     pixx;

short   xtile,ytile;
short   xtilestep,ytilestep;
int32_t    xintercept,yintercept;
word    xstep,ystep;
word    xspot,yspot;
int     texdelta;

word horizwall[MAXWALLTILES],vertwall[MAXWALLTILES];


/*
============================================================================

                           3 - D  DEFINITIONS

============================================================================
*/

/*
========================
=
= TransformActor
=
= Takes paramaters:
=   gx,gy               : globalx/globaly of point
=
= globals:
=   viewx,viewy         : point of view
=   viewcos,viewsin     : sin/cos of viewangle
=   scale               : conversion from global value to screen value
=
= sets:
=   screenx,transx,transy,screenheight: projected edge location and size
=
========================
*/


//
// transform actor
//
void TransformActor (objtype *ob)
{
    fixed gx,gy,gxt,gyt,nx,ny;

//
// translate point to view centered coordinates
//
    gx = ob->x-viewx;
    gy = ob->y-viewy;

//
// calculate newx
//
    gxt = FixedMul(gx,viewcos);
    gyt = FixedMul(gy,viewsin);
    nx = gxt-gyt-ACTORSIZE;         // fudge the shape forward a bit, because
                                    // the midpoint could put parts of the shape
                                    // into an adjacent wall

//
// calculate newy
//
    gxt = FixedMul(gx,viewsin);
    gyt = FixedMul(gy,viewcos);
    ny = gyt+gxt;

//
// calculate perspective ratio
//
    ob->transx = nx;
    ob->transy = ny;

    if (nx<MINDIST)                 // too close, don't overflow the divide
    {
        ob->viewheight = 0;
        return;
    }

    ob->viewx = (word)(centerx + ny*scale/nx);

//
// calculate height (heightnumerator/(nx>>8))
//
    ob->viewheight = (word)(heightnumerator/(nx>>8));
}

//==========================================================================

/*
========================
=
= TransformTile
=
= Takes paramaters:
=   tx,ty               : tile the object is centered in
=
= globals:
=   viewx,viewy         : point of view
=   viewcos,viewsin     : sin/cos of viewangle
=   scale               : conversion from global value to screen value
=
= sets:
=   screenx,transx,transy,screenheight: projected edge location and size
=
= Returns true if the tile is withing getting distance
=
========================
*/

boolean TransformTile (int tx, int ty, short *dispx, short *dispheight)
{
    fixed gx,gy,gxt,gyt,nx,ny;

//
// translate point to view centered coordinates
//
    gx = ((int32_t)tx<<TILESHIFT)+0x8000-viewx;
    gy = ((int32_t)ty<<TILESHIFT)+0x8000-viewy;

//
// calculate newx
//
    gxt = FixedMul(gx,viewcos);
    gyt = FixedMul(gy,viewsin);
    nx = gxt-gyt-0x2000;            // 0x2000 is size of object

//
// calculate newy
//
    gxt = FixedMul(gx,viewsin);
    gyt = FixedMul(gy,viewcos);
    ny = gyt+gxt;


//
// calculate height / perspective ratio
//
    if (nx<MINDIST)                 // too close, don't overflow the divide
        *dispheight = 0;
    else
    {
        *dispx = (short)(centerx + ny*scale/nx);
        *dispheight = (short)(heightnumerator/(nx>>8));
    }

//
// see if it should be grabbed
//
    if (nx<TILEGLOBAL && ny>-TILEGLOBAL/2 && ny<TILEGLOBAL/2)
        return true;
    else
        return false;
}

//==========================================================================

/*
====================
=
= CalcHeight
=
= Calculates the height of xintercept,yintercept from viewx,viewy
=
====================
*/

int CalcHeight()
{
    fixed z = FixedMul(xintercept - viewx, viewcos)
        - FixedMul(yintercept - viewy, viewsin);
    if(z < MINDIST) z = MINDIST;
    int height = heightnumerator / (z >> 8);
    if(height < min_wallheight) min_wallheight = height;
    return height;
}

//==========================================================================

/*
===================
=
= ScalePost
=
===================
*/

byte *postsource;
int postx;
int postwidth;

void ScalePost()
{
    int ywcount, yoffs, yw, yd, yendoffs;
    byte col;

#ifdef USE_SHADING
    byte *curshades = shadetable[GetShade(wallheight[postx])];
#endif

    ywcount = yd = wallheight[postx] >> 3;
    if(yd <= 0) yd = 100;

    yoffs = (viewheight / 2 - ywcount) * vbufPitch;
    if(yoffs < 0) yoffs = 0;
    yoffs += postx;

    yendoffs = viewheight / 2 + ywcount - 1;
    yw=TEXTURESIZE-1;

    while(yendoffs >= viewheight)
    {
        ywcount -= TEXTURESIZE/2;
        while(ywcount <= 0)
        {
            ywcount += yd;
            yw--;
        }
        yendoffs--;
    }
    if(yw < 0) return;

#ifdef USE_SHADING
    col = curshades[postsource[yw]];
#else
    col = postsource[yw];
#endif
    yendoffs = yendoffs * vbufPitch + postx;
    while(yoffs <= yendoffs)
    {
        vbuf[yendoffs] = col;
        ywcount -= TEXTURESIZE/2;
        if(ywcount <= 0)
        {
            do
            {
                ywcount += yd;
                yw--;
            }
            while(ywcount <= 0);
            if(yw < 0) break;
#ifdef USE_SHADING
            col = curshades[postsource[yw]];
#else
            col = postsource[yw];
#endif
        }
        yendoffs -= vbufPitch;
    }
}

void GlobalScalePost(byte *vidbuf, unsigned pitch)
{
    vbuf = vidbuf;
    vbufPitch = pitch;
    ScalePost();
}

/*******************
**
** ApplyPatch
**
********************/
byte   patchedwall[PMPageSize];
int      lastpatchnum;

#define FIRSTPATCHOBJ    2
#define NUMPATCHES       18   // This is the number of patch pairs (light/dark) in use.
#define PATCHSTART       DOORWALL-NUMPATCHES*2

unsigned GetPatchNum(boolean useXspot)
{
    // work off xspot/yspot, instead of xtile/ytile
    unsigned spot, x, y;
    spot = useXspot ? xspot : yspot;
    y = (unsigned)((spot)%MAPSIZE);
    x = (unsigned)((spot+(y == 63 ? 0 : 1))/MAPSIZE);
   
    spot = MAPSPOT(x,y,2);
// Restrict patch objects to the given range.
   if (spot < FIRSTPATCHOBJ || spot >= FIRSTPATCHOBJ+NUMPATCHES )
      return 0;
   else
      return spot;

}

boolean ApplyPatch(int wallpic, boolean useXspot)
{
      byte   *scan;
      int    pixel;
      unsigned spot;

   spot = GetPatchNum(useXspot);
   if (spot > 0 && wallpic < DOORWALL) // Don't do door-related textures.
   {
      // Keep track of the patch in memory.
      //   The wall is already stored elsewhere.
      lastpatchnum = spot;
      /*switch (lastpatchnum)
      {
          case 33:
          case 36:
            spot+=(frameon>>3)%3;
            break;
      }*/
      spot -= FIRSTPATCHOBJ;
      // Adjust for light/dark patch sides
      spot = spot * 2 + (lastside?1:0);
      // Load the wall into the byte array first.
         scan = PM_GetTexture(wallpic);
         for (pixel=0; pixel<PMPageSize; pixel++,scan++) 
             patchedwall[pixel] = *scan;
      // Now add the patch. This MUST be done separate from wallpic's page.
         scan = PM_GetTexture(PATCHSTART+spot);
         for (pixel=0; pixel<PMPageSize; pixel++,scan++) 
            if (*scan < 255)
             patchedwall[pixel] = *scan;
      return true;
   }
   lastpatchnum = 0;
   return false;
}
/*
====================
=
= HitVertWall
=
= tilehit bit 7 is 0, because it's not a door tile
= if bit 6 is 1 and the adjacent tile is a door tile, use door side pic
=
====================
*/

void HitVertWall (void)
{
    int wallpic;
    int texture;
    
    ytile = (short)(yintercept>>TILESHIFT);
    texture = ((yintercept+texdelta)>>TEXTUREFROMFIXEDSHIFT)&TEXTUREMASK;
    if (xtilestep == -1)
    {
        texture = TEXTUREMASK-texture;
        xintercept += TILEGLOBAL;
    }

#ifdef WALLPATCH
    if(lastside==1 && lastintercept==xtile && lasttilehit==tilehit &&
        !(lasttilehit & BIT_WALL) && GetPatchNum(true)==lastpatchnum)
#else
    if(lastside==1 && lastintercept==xtile && lasttilehit==tilehit && !(lasttilehit & BIT_WALL))
#endif
    {
        if((pixx&3) && texture == lasttexture)
        {
            ScalePost();
            postx = pixx;
            wallheight[pixx] = wallheight[pixx-1];
            return;
        }
        ScalePost();
        wallheight[pixx] = CalcHeight();

        postsource+=texture-lasttexture;
        postwidth=1;
        postx=pixx;
        lasttexture=texture;
        return;
    }

    if(lastside!=-1) ScalePost();

    lastside=1;
    lastintercept=xtile;
    lasttilehit=tilehit;
    lasttexture=texture;
    wallheight[pixx] = CalcHeight();
    postx = pixx;
    postwidth = 1;

    // Animated walls
    /*switch (lasttilehit)
    {
        case 2:
        case 42:
            tilehit+=(frameon>>3)%4;
            break;
    }*/

    if (tilehit & BIT_WALL)
    {                                                               // check for adjacent doors
        if (tilemap[xtile-xtilestep][ytile]&BIT_DOOR)
        {
            switch (doorobjlist[tilemap[xtile-xtilestep][ytile] & LAST_DOORNUM].lock)
            {
                case dr_lock1:
                    wallpic = DOORWALL+9;
                    break;
                case dr_lock2:
                    wallpic = DOORWALL+11;
                    break;
                case dr_lock3:
                    wallpic = DOORWALL+13;
                    break;
                case dr_lock4:
                case dr_normal:
                case dr_elevator:
                case dr_green:
                case dr_brown:
                default:
                    wallpic = DOORWALL+3;
                    break;
            }
    }
    else
        wallpic = vertwall[tilehit & ~BIT_WALL];
    }
    else
        wallpic = vertwall[tilehit];
#ifdef WALLPATCH
    if (ApplyPatch (wallpic, true))
      postsource = patchedwall+texture;
    else
#endif
      postsource = PM_GetTexture(wallpic) + texture;
}


/*
====================
=
= HitHorizWall
=
= tilehit bit 7 is 0, because it's not a door tile
= if bit 6 is 1 and the adjacent tile is a door tile, use door side pic
=
====================
*/

void HitHorizWall (void)
{
    int wallpic;
    int texture;
    
    xtile = (short)(xintercept>>TILESHIFT);
    texture = ((xintercept+texdelta)>>TEXTUREFROMFIXEDSHIFT)&TEXTUREMASK;
    if (ytilestep == -1)
        yintercept += TILEGLOBAL;
    else
        texture = TEXTUREMASK-texture;
    
#ifdef WALLPATCH
    if(lastside==0 && lastintercept==ytile && lasttilehit==tilehit &&
        !(lasttilehit & BIT_WALL) && GetPatchNum(false)==lastpatchnum)
#else
    if(lastside==0 && lastintercept==ytile && lasttilehit==tilehit && !(lasttilehit & BIT_WALL))
#endif
    {
        if((pixx&3) && texture == lasttexture)
        {
            ScalePost();
            postx=pixx;
            wallheight[pixx] = wallheight[pixx-1];
            return;
        }
        ScalePost();
        wallheight[pixx] = CalcHeight();
        postsource+=texture-lasttexture;
        postwidth=1;
        postx=pixx;
        lasttexture=texture;
        return;
    }

    if(lastside!=-1) ScalePost();

    lastside=0;
    lastintercept=ytile;
    lasttilehit=tilehit;
    lasttexture=texture;
    wallheight[pixx] = CalcHeight();
    postx = pixx;
    postwidth = 1;
    
    // Animated walls
    /*switch (lasttilehit)
    {
        case 2:
        case 42:
            tilehit+=(frameon>>3)%4;
            break;
    }*/

    if (tilehit & BIT_WALL)
    {                                                  // check for adjacent doors
        if (tilemap[xtile][ytile-ytilestep]&BIT_DOOR)
        {
            switch (doorobjlist[tilemap[xtile][ytile-ytilestep] & LAST_DOORNUM].lock)
            {
                case dr_lock1:
                    wallpic = DOORWALL+8;
                    break;
                case dr_lock2:
                    wallpic = DOORWALL+10;
                    break;
                case dr_lock3:
                    wallpic = DOORWALL+12;
                    break;
                case dr_normal:
                case dr_elevator:
                case dr_lock4:
                case dr_green:
                case dr_brown:
                default:
                    wallpic = DOORWALL+2;
                    break;
            }
        }
        else
            wallpic = horizwall[tilehit & ~BIT_WALL];
    }
    else
        wallpic = horizwall[tilehit];
    
#ifdef WALLPATCH
    if (ApplyPatch (wallpic, false))
        postsource = patchedwall+texture;
    else
#endif
        postsource = PM_GetTexture(wallpic) + texture;
}

//==========================================================================

/*
====================
=
= HitHorizDoor
=
====================
*/

void HitHorizDoor (void)
{
    int doorpage;
    int doornum;
    int texture;

    doornum = tilehit&LAST_DOORNUM;
    texture = ((xintercept-doorposition[doornum])>>TEXTUREFROMFIXEDSHIFT)&TEXTUREMASK;

    if(lasttilehit==tilehit)
    {
        if((pixx&3) && texture == lasttexture)
        {
            ScalePost();
            postx=pixx;
            wallheight[pixx] = wallheight[pixx-1];
            return;
        }
        ScalePost();
        wallheight[pixx] = CalcHeight();
        postsource+=texture-lasttexture;
        postwidth=1;
        postx=pixx;
        lasttexture=texture;
        return;
    }

    if(lastside!=-1) ScalePost();

    lastside=2;
    lasttilehit=tilehit;
    lasttexture=texture;
    wallheight[pixx] = CalcHeight();
    postx = pixx;
    postwidth = 1;

    switch(doorobjlist[doornum].lock)
    {
        case dr_normal:
            doorpage = DOORWALL;
            break;
        case dr_lock1:
        case dr_lock2:
        case dr_lock3:
        case dr_brown:
            doorpage = DOORWALL+6;
            break;
        case dr_lock4:
            doorpage = DOORWALL+16;
            break;
        case dr_green:
            doorpage = DOORWALL+18;
            break;
        case dr_elevator:
            doorpage = DOORWALL+4;
            break;
    }

    postsource = PM_GetTexture(doorpage) + texture;
}

//==========================================================================

/*
====================
=
= HitVertDoor
=
====================
*/

void HitVertDoor (void)
{
    int doorpage;
    int doornum;
    int texture;

    doornum = tilehit&LAST_DOORNUM;
    texture = ((yintercept-doorposition[doornum])>>TEXTUREFROMFIXEDSHIFT)&TEXTUREMASK;

    if(lasttilehit==tilehit)
    {
        if((pixx&3) && texture == lasttexture)
        {
            ScalePost();
            postx=pixx;
            wallheight[pixx] = wallheight[pixx-1];
            return;
        }
        ScalePost();
        wallheight[pixx] = CalcHeight();
        postsource+=texture-lasttexture;
        postwidth=1;
        postx=pixx;
        lasttexture=texture;
        return;
    }

    if(lastside!=-1) ScalePost();

    lastside=2;
    lasttilehit=tilehit;
    lasttexture=texture;
    wallheight[pixx] = CalcHeight();
    postx = pixx;
    postwidth = 1;

    switch(doorobjlist[doornum].lock)
    {
        case dr_normal:
            doorpage = DOORWALL+1;
            break;
        case dr_lock1:
        case dr_lock2:
        case dr_lock3:
        case dr_brown:
            doorpage = DOORWALL+6;
            break;
        case dr_lock4:
            doorpage = DOORWALL+16;
            break;
        case dr_green:
            doorpage = DOORWALL+18;
            break;
        case dr_elevator:
            doorpage = DOORWALL+4;
            break;
    }

    postsource = PM_GetTexture(doorpage) + texture;
}

//==========================================================================

#define HitHorizBorder HitHorizWall
#define HitVertBorder HitVertWall

//==========================================================================


/*
=====================
=
= VGAClearScreen
=
=====================
*/

void VGAClearScreen (void)
{
    byte ceiling=0; // black 0x00

    int y;
    byte *ptr = vbuf;
#ifdef USE_SHADING
    for(y = 0; y < viewheight / 2; y++, ptr += vbufPitch)
        memset(ptr, shadetable[GetShade((viewheight / 2 - y) << 3)][ceiling], viewwidth);
    for(; y < viewheight; y++, ptr += vbufPitch)
        memset(ptr, shadetable[GetShade((y - viewheight / 2) << 3)][0x00], viewwidth);
#else
    for(y = 0; y < viewheight / 2; y++, ptr += vbufPitch)
        memset(ptr, ceiling, viewwidth);
    for(; y < viewheight; y++, ptr += vbufPitch)
        memset(ptr, 0x19, viewwidth);
#endif
}

//==========================================================================

/*
=====================
=
= CalcRotate
=
=====================
*/

int CalcRotate (objtype *ob)
{
    int angle, viewangle;

    // this isn't exactly correct, as it should vary by a trig value,
    // but it is close enough with only eight rotations

    viewangle = player->angle + (centerx - ob->viewx)/8;

    switch (ob->obclass)
    {
        case rocketobj:
        case cyberrocketobj:
        case revballobj:
        case bossballobj:
        case fatballobj:
            angle = (viewangle-180)- ob->angle;
            break;
        default:
            angle = (viewangle-180)- dirangle[ob->dir];
            break;
    }

    angle+=ANGLES/16;
    while (angle>=ANGLES)
        angle-=ANGLES;
    while (angle<0)
        angle+=ANGLES;

    if (ob->state->rotate == 2)             // 2 rotation pain frame
        return 0;               // pain with shooting frame bugfix

    return angle/(ANGLES/8);
}

void ScaleShape (int xcenter, int shapenum, unsigned height, uint32_t flags)
{
    t_compshape *shape;
    unsigned scale,pixheight;
    unsigned starty,endy;
    word *cmdptr;
    byte *cline;
    byte *line;
    byte *vmem;
    int actx,i,upperedge;
    short newstart;
    int scrstarty,screndy,lpix,rpix,pixcnt,ycnt;
    unsigned j;
    byte col;

#ifdef USE_SHADING
    byte *curshades;
    if(flags & FL_FULLBRIGHT)
        curshades = shadetable[0];
    else
        curshades = shadetable[GetShade(height)];
#endif
    if (shapenum == 0)
        return;
    shape = (t_compshape *) PM_GetSprite(shapenum);

    scale=height>>3;                 // low three bits are fractional
    if(!scale) return;   // too close or far away
    
    // Mech column, large torches, lamp posts, hanging gibs, and the large tree
    if (shapenum == SPR_STAT_48 || (shapenum >= SPR_STAT_54 && shapenum <= SPR_STAT_57) || (shapenum >= SPR_STAT3_10 && shapenum <= SPR_STAT3_37)
        || (shapenum >= SPR_STAT3_53 && shapenum <= SPR_STAT3_58)
        || shapenum == SPR_STAT3_64) {
        pixheight = scale;
        actx=xcenter-scale;
    }
    else {
        pixheight = scale * 2;
        actx=xcenter-scale*2;
    }
    if (shapenum >= SPR_ROCKET_1 && shapenum <= SPR_BOOM_3) // and explosion too
        upperedge=viewheight/2-scale*2;
    else if (shapenum == SPR_STAT_48 || (shapenum >= SPR_STAT_54 && shapenum <= SPR_STAT_57) ||  (shapenum >= SPR_STAT3_10 && shapenum <= SPR_STAT3_37)
        || (shapenum >= SPR_STAT3_53 && shapenum <= SPR_STAT3_58)
        || shapenum == SPR_STAT3_64) {
        upperedge = viewheight / 2 - scale;
    }
    else
        upperedge=viewheight/2-scale*3;

    cmdptr=(word *) shape->dataofs;

    for(i=shape->leftpix,pixcnt=i*pixheight,rpix=(pixcnt>>6)+actx;i<=shape->rightpix;i++,cmdptr++)
    {
        lpix=rpix;
        if(lpix>=viewwidth) break;
        pixcnt+=pixheight;
        rpix=(pixcnt>>6)+actx;
        if(lpix!=rpix && rpix>0)
        {
            if(lpix<0) lpix=0;
            if(rpix>viewwidth) rpix=viewwidth,i=shape->rightpix+1;
            cline=(byte *)shape + *cmdptr;
            while(lpix<rpix)
            {
                if(wallheight[lpix]<=(int)height)
                {
                    line=cline;
                    while((endy = READWORD(line)) != 0)
                    {
                        endy >>= 1;
                        newstart = READWORD(line);
                        starty = READWORD(line) >> 1;
                        j=starty;
                        ycnt=j*pixheight;
                        screndy=(ycnt>>6)+upperedge;
                        if(screndy<0) vmem=vbuf+lpix;
                        else vmem=vbuf+screndy*vbufPitch+lpix;
                        for(;j<endy;j++)
                        {
                            scrstarty=screndy;
                            ycnt+=pixheight;
                            screndy=(ycnt>>6)+upperedge;
                            if(scrstarty!=screndy && screndy>0)
                            {
#ifdef USE_SHADING
                                col=curshades[((byte *)shape)[newstart+j]];
#else
                                col=((byte *)shape)[newstart+j];
#endif
                                if(scrstarty<0) scrstarty=0;
                                if(screndy>viewheight) screndy=viewheight,j=endy;

                                while(scrstarty<screndy)
                                {
                                    *vmem=col;
                                    vmem+=vbufPitch;
                                    scrstarty++;
                                }
                            }
                        }
                    }
                }
                lpix++;
            }
        }
    }
}

void SimpleScaleShape (int xcenter, int shapenum, unsigned height, float shapeScale = 1.0f)
{
    t_compshape   *shape;
    unsigned scale,pixheight;
    unsigned starty,endy;
    word *cmdptr;
    byte *cline;
    byte *line;
    int actx,i,upperedge;
    short newstart;
    int scrstarty,screndy,lpix,rpix,pixcnt,ycnt;
    unsigned j;
    byte col;
    byte *vmem;
    int inviscolor[2] = {0,255};

    shape = (t_compshape *) PM_GetSprite(shapenum);
    
    scale=(int)(64*(screenHeight/200) * shapeScale);
    pixheight=scale*SPRITESCALEFACTOR;
    actx=xcenter-scale;

    if (viewsize==20)
    {
        if (gamestate.weapon == wp_chaingun || gamestate.weapon == wp_launcher ||
            gamestate.weapon == wp_plasmagun || gamestate.weapon == wp_bfggun ||
            gamestate.weapon == wp_chainsaw)
            upperedge = screenHeight-((126-height)*(screenHeight/200));
        else
            upperedge=screenHeight-((int)(150 * shapeScale-height)*(screenHeight/200));
    }
    else
    {
        if (gamestate.weapon == wp_chaingun || gamestate.weapon == wp_launcher)
            upperedge=screenHeight-((104-height)*(screenHeight/200));
        else
            upperedge=screenHeight-((int)(128 * shapeScale-height)*(screenHeight/200));
    }

    cmdptr=shape->dataofs;

    for(i=shape->leftpix,pixcnt=i*pixheight,rpix=(pixcnt>>6)+actx;i<=shape->rightpix;i++,cmdptr++)
    {
        lpix=rpix;
        if(lpix>=viewwidth) break;
        pixcnt+=pixheight;
        rpix=(pixcnt>>6)+actx;
        if(lpix!=rpix && rpix>0)
        {
            if(lpix<0) lpix=0;
            if(rpix>viewwidth) rpix=viewwidth,i=shape->rightpix+1;
            cline = (byte *)shape + *cmdptr;
            while(lpix<rpix)
            {
                line=cline;
                while((endy = READWORD(line)) != 0)
                {
                    endy >>= 1;
                    newstart = READWORD(line);
                    starty = READWORD(line) >> 1;
                    j=starty;
                    ycnt=j*pixheight;
                    screndy=(ycnt>>6)+upperedge;
                    if(screndy<0) vmem=vbuf+lpix;
                    else vmem=vbuf+screndy*vbufPitch+lpix;
                    for(;j<endy;j++)
                    {
                        scrstarty=screndy;
                        ycnt+=pixheight;
                        screndy=(ycnt>>6)+upperedge;
                        if(scrstarty!=screndy && screndy>0)
                        {
                            if (gamestate.startinvistrigger)
                                col=inviscolor[US_RndT()%2];
                            else
                                col=((byte *)shape)[newstart+j];
                            if(scrstarty<0) scrstarty=0;
                            if(screndy>viewheight) screndy=viewheight,j=endy;

                            while(scrstarty<screndy)
                            {
                                if (col != 255)
                                    *vmem=col;
                                vmem+=vbufPitch;
                                scrstarty++;
                            }
                        }
                    }
                }
                lpix++;
            }
        }
    }
}

/*
=====================
=
= DrawScaleds
=
= Draws all objects that are visable
=
=====================
*/

#define MAXVISABLE 250

typedef struct
{
    short      viewx,
               viewheight,
               shapenum;
    short      flags;          // this must be changed to uint32_t, when you
                               // you need more than 16-flags for drawing
#ifdef USE_DIR3DSPR
    statobj_t *transsprite;
#endif
} visobj_t;

visobj_t vislist[MAXVISABLE];
visobj_t *visptr,*visstep,*farthest;

void DrawScaleds (void)
{
    int      i,least,numvisable,height;
    uint16_t     *tilespot;
    byte     *visspot;
    unsigned spotloc;

    statobj_t *statptr;
    objtype   *obj;

    visptr = &vislist[0];

//
// place static objects
//
    for (statptr = &statobjlist[0] ; statptr !=laststatobj ; statptr++)
    {
        if ((visptr->shapenum = statptr->shapenum) == -1)
            continue;                                               // object has been deleted

        if (!*statptr->visspot)
            continue;                                               // not visable
        switch (statptr->shapenum)
        {
            case SPR_STAT_0:
            case SPR_STAT_2:
            case SPR_STAT_4:
            case SPR_STAT_21:
            case SPR_STAT_23:
            case SPR_STAT3_0:
            case SPR_STAT3_8:
            case SPR_STAT3_42:
            case SPR_STAT_51:
                visptr->shapenum+=(frameon>>4)%2;
                break;
            case SPR_STAT_43:
            case SPR_STAT3_2:
            case SPR_STAT3_5:
                visptr->shapenum+=(frameon>>4)%3;
                break;
            case SPR_STAT_54:
            case SPR_STAT3_10:
            case SPR_STAT3_14:
            case SPR_STAT3_18:
            case SPR_STAT3_22:
            case SPR_STAT3_26:
            case SPR_STAT3_30:
            case SPR_STAT3_34:
            case SPR_HELLORB:
                visptr->shapenum+=(frameon>>3)%4;
                break;
            case SPR_STAT_25:
            case SPR_STAT_29:
            case SPR_STAT_37:
            case SPR_STAT_58:
            case SPR_STAT3_38:
            case SPR_STAT3_44:
                visptr->shapenum += abs(3 - (frameon>>3)%6);
                break;
        }
        if (TransformTile (statptr->tilex,statptr->tiley,
            &visptr->viewx,&visptr->viewheight) && statptr->flags & FL_BONUS)
        {
            GetBonus (statptr);
            if(statptr->shapenum == -1)
                continue;                                           // object has been taken
        }

        if (!visptr->viewheight)
            continue;                                               // to close to the object

#ifdef USE_DIR3DSPR
        if(statptr->flags & FL_DIR_MASK)
            visptr->transsprite=statptr;
        else
            visptr->transsprite=NULL;
#endif

        if (visptr < &vislist[MAXVISABLE-1])    // don't let it overflow
        {
            visptr->flags = (short) statptr->flags;
            visptr++;
        }
    }

//
// place active objects
//
    for (obj = player->next;obj;obj=obj->next)
    {
        if ((visptr->shapenum = obj->state->shapenum)==0)
            continue;                                               // no shape

        spotloc = (obj->tilex<<mapshift)+obj->tiley;   // optimize: keep in struct?
        visspot = &spotvis[0][0]+spotloc;
        tilespot = &tilemap[0][0]+spotloc;

        //
        // could be in any of the nine surrounding tiles
        //
        if (*visspot
            || ( *(visspot-1) && !*(tilespot-1) )
            || ( *(visspot+1) && !*(tilespot+1) )
            || ( *(visspot-65) && !*(tilespot-65) )
            || ( *(visspot-64) && !*(tilespot-64) )
            || ( *(visspot-63) && !*(tilespot-63) )
            || ( *(visspot+65) && !*(tilespot+65) )
            || ( *(visspot+64) && !*(tilespot+64) )
            || ( *(visspot+63) && !*(tilespot+63) ) )
        {
            obj->active = ac_yes;
            TransformActor (obj);
            if (!obj->viewheight)
                continue;                                               // too close or far away

            visptr->viewx = obj->viewx;
            visptr->viewheight = obj->viewheight;
            if (visptr->shapenum == -1)
                visptr->shapenum = obj->temp1;  // special shape

            if (obj->state->rotate)
                visptr->shapenum += CalcRotate (obj);

            if (visptr < &vislist[MAXVISABLE-1])    // don't let it overflow
            {
                visptr->flags = (short) obj->flags;
#ifdef USE_DIR3DSPR
                visptr->transsprite = NULL;
#endif
                visptr++;
            }
            obj->flags |= FL_VISABLE;
        }
        else
            obj->flags &= ~FL_VISABLE;
    }

//
// draw from back to front
//
    numvisable = (int) (visptr-&vislist[0]);

    if (!numvisable)
        return;                                                                 // no visable objects

    for (i = 0; i<numvisable; i++)
    {
        least = 32000;
        for (visstep=&vislist[0] ; visstep<visptr ; visstep++)
        {
            height = visstep->viewheight;
            if (height < least)
            {
                least = height;
                farthest = visstep;
            }
        }
        //
        // draw farthest
        //
#ifdef USE_DIR3DSPR
        if(farthest->transsprite)
            Scale3DShape(vbuf, vbufPitch, farthest->transsprite);
        else
#endif
            ScaleShape(farthest->viewx, farthest->shapenum, farthest->viewheight, farthest->flags);

        farthest->viewheight = 32000;
    }
}

//==========================================================================

/*
==============
=
= DrawPlayerWeapon
=
= Draw the player's hands
=
==============
*/

int weaponscale[NUMWEAPONS] = {SPR_FISTS_READY,SPR_CHAINSAW_IDLE1,SPR_PISTOL_READY,SPR_SHOTGUN_READY,
                    SPR_SUPERSHOT_READY,SPR_CHAINGUN_READY,SPR_LAUNCHER_READY,
                    SPR_PLASMAGUN_READY,SPR_BFGGUN_READY, SPR_HELLORB_READY};

void DrawPlayerWeapon (void)
{
    int shapenum;
    
    shapenum = weaponscale[gamestate.weapon]+gamestate.weaponframe;
    if (shapenum == SPR_CHAINSAW_IDLE1) {
        shapenum+=(frameon>>3)%2;
    }

    if (shapenum == SPR_HELLORB_READY) {
        shapenum += (frameon >> 3) % 4;
    }

    if (shapenum==SPR_CHAINSAW_IDLE2)
    {
        SD_PlaySound (DSSAWIDL, false);
    }

    if (gamestate.weapon != -1)
    {
        switch (shapenum)
        {
            case SPR_FISTS_READY:
                SimpleScaleShape (viewwidth-(84*(screenHeight/200))+gamestate.bobber2,SPR_FISTS_READY,0+gamestate.bobber+gamestate.weapchange);
                SimpleScaleShape ((44*(screenHeight/200))+gamestate.bobber2,SPR_FISTS_READY2,0+gamestate.bobber+gamestate.weapchange);
                break;
            case SPR_FISTS_ATK1:
                SimpleScaleShape ((64*(screenHeight/200)),SPR_FISTS_ATK1,0);
                break;
            case SPR_FISTS_ATK2:
                SimpleScaleShape ((83*(screenHeight/200)),SPR_FISTS_ATK2,0);
                SimpleScaleShape ((64*(screenHeight/200)),SPR_FISTS_ATK2A,0);
                break;
            case SPR_SHOTGUN_RLD1:
            case SPR_SUPERSHOT_RLD1:
                SimpleScaleShape(viewwidth/2-(20*(screenHeight/200)),shapenum,0);
                break;
            case SPR_SHOTGUN_RLD2:
            case SPR_SUPERSHOT_RLD2:
            case SPR_SUPERSHOT_RLD6:
                SimpleScaleShape(viewwidth/2-(30*(screenHeight/200)),shapenum,0);
                break;
            case SPR_SHOTGUN_RLD3:
            case SPR_SUPERSHOT_RLD3:
                SimpleScaleShape(viewwidth/2-(60*(screenHeight/200)),shapenum,0);
                break;
            case SPR_SUPERSHOT_RLD4:
                SimpleScaleShape(viewwidth/2-(60*(screenHeight/200)),shapenum,0);
                SimpleScaleShape(viewwidth/2+(60*(screenHeight/200)),SPR_SUPERSHOT_RLD4A,0);
                break;
            case SPR_BFGGUN_READY:
                SimpleScaleShape(viewwidth/2-(84*(screenHeight/200))+gamestate.bobber2,shapenum,32+gamestate.bobber+gamestate.weapchange);
                SimpleScaleShape(viewwidth/2+(44*(screenHeight/200))+gamestate.bobber2,SPR_BFGGUN_READYA,32+gamestate.bobber+gamestate.weapchange);
                break;
            case SPR_BFGGUN_ATK1:
                SimpleScaleShape(viewwidth/2-(64*(screenHeight/200)),shapenum,32);
                SimpleScaleShape(viewwidth/2+(64*(screenHeight/200)),SPR_BFGGUN_ATK1A,32);
                break;
            case SPR_BFGGUN_ATK2:
                SimpleScaleShape(viewwidth/2-(64*(screenHeight/200)),shapenum,32);
                SimpleScaleShape(viewwidth/2+(64*(screenHeight/200)),SPR_BFGGUN_ATK2A,32);
                break;
            case SPR_BFGGUN_ATK3:
                SimpleScaleShape(viewwidth/2-(64*(screenHeight/200)),shapenum,32);
                SimpleScaleShape(viewwidth/2+(64*(screenHeight/200)),SPR_BFGGUN_ATK3A,32);
                break;
            case SPR_CHAINSAW_IDLE1:
                SimpleScaleShape(viewwidth/2-(84*(screenHeight/200))+gamestate.bobber2,shapenum,0+gamestate.bobber+gamestate.weapchange);
                SimpleScaleShape(viewwidth/2+(44*(screenHeight/200))+gamestate.bobber2,SPR_CHAINSAW_IDLE1A,0+gamestate.bobber+gamestate.weapchange);
                shapenum+=(frameon>>3)%2;
                break;
            case SPR_CHAINSAW_IDLE2:
                SimpleScaleShape(viewwidth/2-(84*(screenHeight/200))+gamestate.bobber2,shapenum,0+gamestate.bobber+gamestate.weapchange);
                SimpleScaleShape(viewwidth/2+(44*(screenHeight/200))+gamestate.bobber2,SPR_CHAINSAW_IDLE2A,0+gamestate.bobber+gamestate.weapchange);
                break;
            case SPR_CHAINSAW_ATK1:
                SimpleScaleShape(viewwidth/2-(64*(screenHeight/200)),shapenum,0);
                SimpleScaleShape(viewwidth/2+(64*(screenHeight/200)),SPR_CHAINSAW_ATK1A,0);
                break;
            case SPR_CHAINSAW_ATK2:
                SimpleScaleShape(viewwidth/2-(64*(screenHeight/200)),shapenum,0);
                SimpleScaleShape(viewwidth/2+(64*(screenHeight/200)),SPR_CHAINSAW_ATK2A,0);
                break;
            case SPR_PISTOL_READY:
            case SPR_SHOTGUN_READY:
            case SPR_SUPERSHOT_READY:
            case SPR_CHAINGUN_READY:
            case SPR_LAUNCHER_READY:
            case SPR_PLASMAGUN_READY:
                SimpleScaleShape(viewwidth/2-(20*(screenHeight/200))+gamestate.bobber2,shapenum,0+gamestate.bobber+gamestate.weapchange);
                break;
            case SPR_HELLORB_READY:
            case SPR_HELLORB_READY2:
            case SPR_HELLORB_READY3:
            case SPR_HELLORB_READY4:
            case SPR_HELLORB_ATK1:
            case SPR_HELLORB_ATK2:
            case SPR_HELLORB_ATK3:
            case SPR_HELLORB_ATK4:
            case SPR_HELLORB_ATK5:
            case SPR_HELLORB_ATK6:
            case SPR_HELLORB_ATK7:
            case SPR_HELLORB_ATK8:
                SimpleScaleShape((int)(viewwidth-184*1.5),shapenum,0, 1.5f);
                break;
            default:
                SimpleScaleShape(viewwidth/2,shapenum,0);
        }
    }

    if (demorecord || demoplayback)
        SimpleScaleShape(viewwidth/2,SPR_DEMO,viewheight+1);
}


//==========================================================================


/*
=====================
=
= CalcTics
=
=====================
*/

void CalcTics (void)
{
//
// calculate tics since last refresh for adaptive timing
//
    if (lasttimecount > (int32_t) GetTimeCount())
        lasttimecount = GetTimeCount();    // if the game was paused a LONG time

    uint32_t curtime = SDL_GetTicks();
    tics = (curtime * 7) / 100 - lasttimecount;
    if(!tics)
    {
        // wait until end of current tic
        SDL_Delay(((lasttimecount + 1) * 100) / 7 - curtime);
        tics = 1;
    }

    lasttimecount += tics;

    if (tics>MAXTICS)
        tics = MAXTICS;
}


//==========================================================================

void AsmRefresh()
{
    int32_t xstep,ystep;
    longword xpartial,ypartial;
    boolean playerInPushwallBackTile = tilemap[focaltx][focalty] == BIT_WALL;

    for(pixx=0;pixx<viewwidth;pixx++)
    {
        short angl=midangle+pixelangle[pixx];
        if(angl<0) angl+=FINEANGLES;
        if(angl>=3600) angl-=FINEANGLES;
        if(angl<900)
        {
            xtilestep=1;
            ytilestep=-1;
            xstep=finetangent[900-1-angl];
            ystep=-finetangent[angl];
            xpartial=xpartialup;
            ypartial=ypartialdown;
        }
        else if(angl<1800)
        {
            xtilestep=-1;
            ytilestep=-1;
            xstep=-finetangent[angl-900];
            ystep=-finetangent[1800-1-angl];
            xpartial=xpartialdown;
            ypartial=ypartialdown;
        }
        else if(angl<2700)
        {
            xtilestep=-1;
            ytilestep=1;
            xstep=-finetangent[2700-1-angl];
            ystep=finetangent[angl-1800];
            xpartial=xpartialdown;
            ypartial=ypartialup;
        }
        else if(angl<3600)
        {
            xtilestep=1;
            ytilestep=1;
            xstep=finetangent[angl-2700];
            ystep=finetangent[3600-1-angl];
            xpartial=xpartialup;
            ypartial=ypartialup;
        }
        yintercept=FixedMul(ystep,xpartial)+viewy;
        xtile=focaltx+xtilestep;
        xspot=(word)((xtile<<mapshift)+((uint32_t)yintercept>>16));
        xintercept=FixedMul(xstep,ypartial)+viewx;
        ytile=focalty+ytilestep;
        yspot=(word)((((uint32_t)xintercept>>16)<<mapshift)+ytile);
        texdelta=0;

        // Special treatment when player is in back tile of pushwall
        if(playerInPushwallBackTile)
        {
            if(    pwalldir == di_east && xtilestep ==  1
                || pwalldir == di_west && xtilestep == -1)
            {
                int32_t yintbuf = yintercept - ((ystep * (64 - pwallpos)) >> 6);
                if((yintbuf >> 16) == focalty)   // ray hits pushwall back?
                {
                    if(pwalldir == di_east)
                        xintercept = (focaltx << TILESHIFT) + (pwallpos << 10);
                    else
                        xintercept = (focaltx << TILESHIFT) - TILEGLOBAL + ((64 - pwallpos) << 10);
                    yintercept = yintbuf;
                    ytile = (short) (yintercept >> TILESHIFT);
                    tilehit = pwalltile;
                    HitVertWall();
                    continue;
                }
            }
            else if(pwalldir == di_south && ytilestep ==  1
                ||  pwalldir == di_north && ytilestep == -1)
            {
                int32_t xintbuf = xintercept - ((xstep * (64 - pwallpos)) >> 6);
                if((xintbuf >> 16) == focaltx)   // ray hits pushwall back?
                {
                    xintercept = xintbuf;
                    if(pwalldir == di_south)
                        yintercept = (focalty << TILESHIFT) + (pwallpos << 10);
                    else
                        yintercept = (focalty << TILESHIFT) - TILEGLOBAL + ((64 - pwallpos) << 10);
                    xtile = (short) (xintercept >> TILESHIFT);
                    tilehit = pwalltile;
                    HitHorizWall();
                    continue;
                }
            }
        }

        do
        {
            if(ytilestep==-1 && (yintercept>>16)<=ytile) goto horizentry;
            if(ytilestep==1 && (yintercept>>16)>=ytile) goto horizentry;
vertentry:
            if((uint32_t)yintercept>mapheight*65536-1 || (word)xtile>=mapwidth)
            {
                if(xtile<0) xintercept=0, xtile=0;
                else if(xtile>=mapwidth) xintercept=mapwidth<<TILESHIFT, xtile=mapwidth-1;
                else xtile=(short) (xintercept >> TILESHIFT);
                if(yintercept<0) yintercept=0, ytile=0;
                else if(yintercept>=(mapheight<<TILESHIFT)) yintercept=mapheight<<TILESHIFT, ytile=mapheight-1;
                yspot=0xffff;
                tilehit=0;
                HitHorizBorder();
                break;
            }
            if(xspot>=maparea) break;
            tilehit=((uint16_t *)tilemap)[xspot];
            if(tilehit)
            {
                if(tilehit&BIT_DOOR)
                {
                    int32_t yintbuf=yintercept+(ystep>>1);
                    if((yintbuf>>16)!=(yintercept>>16))
                        goto passvert;
                    if((word)yintbuf<doorposition[tilehit&LAST_DOORNUM])
                        goto passvert;
                    yintercept=yintbuf;
                    xintercept=(xtile<<TILESHIFT)|0x8000;
                    ytile = (short) (yintercept >> TILESHIFT);
                    HitVertDoor();
                }
                else
                {
                    if(tilehit==BIT_WALL)
                    {
                        if(pwalldir==di_west || pwalldir==di_east)
                        {
                            int32_t yintbuf;
                            int pwallposnorm;
                            int pwallposinv;
                            if(pwalldir==di_west)
                            {
                                pwallposnorm = 64-pwallpos;
                                pwallposinv = pwallpos;
                            }
                            else
                            {
                                pwallposnorm = pwallpos;
                                pwallposinv = 64-pwallpos;
                            }
                            if(pwalldir == di_east && xtile==pwallx && ((uint32_t)yintercept>>16)==pwally
                                || pwalldir == di_west && !(xtile==pwallx && ((uint32_t)yintercept>>16)==pwally))
                            {
                                yintbuf=yintercept+((ystep*pwallposnorm)>>6);
                                if((yintbuf>>16)!=(yintercept>>16))
                                    goto passvert;

                                xintercept=(xtile<<TILESHIFT)+TILEGLOBAL-(pwallposinv<<10);
                                yintercept=yintbuf;
                                ytile = (short) (yintercept >> TILESHIFT);
                                tilehit=pwalltile;
                                HitVertWall();
                            }
                            else
                            {
                                yintbuf=yintercept+((ystep*pwallposinv)>>6);
                                if((yintbuf>>16)!=(yintercept>>16))
                                    goto passvert;

                                xintercept=(xtile<<TILESHIFT)-(pwallposinv<<10);
                                yintercept=yintbuf;
                                ytile = (short) (yintercept >> TILESHIFT);
                                tilehit=pwalltile;
                                HitVertWall();
                            }
                        }
                        else
                        {
                            int pwallposi = pwallpos;
                            if(pwalldir==di_north) pwallposi = 64-pwallpos;
                            if(pwalldir==di_south && (word)yintercept<(pwallposi<<10)
                                || pwalldir==di_north && (word)yintercept>(pwallposi<<10))
                            {
                                if(((uint32_t)yintercept>>16)==pwally && xtile==pwallx)
                                {
                                    if(pwalldir==di_south && (int32_t)((word)yintercept)+ystep<(pwallposi<<10)
                                            || pwalldir==di_north && (int32_t)((word)yintercept)+ystep>(pwallposi<<10))
                                        goto passvert;

                                    if(pwalldir==di_south)
                                        yintercept=(yintercept&0xffff0000)+(pwallposi<<10);
                                    else
                                        yintercept=(yintercept&0xffff0000)-TILEGLOBAL+(pwallposi<<10);
                                    xintercept=xintercept-((xstep*(64-pwallpos))>>6);
                                    xtile = (short) (xintercept >> TILESHIFT);
                                    tilehit=pwalltile;
                                    HitHorizWall();
                                }
                                else
                                {
                                    texdelta = -(pwallposi<<10);
                                    xintercept=xtile<<TILESHIFT;
                                    ytile = (short) (yintercept >> TILESHIFT);
                                    tilehit=pwalltile;
                                    HitVertWall();
                                }
                            }
                            else
                            {
                                if(((uint32_t)yintercept>>16)==pwally && xtile==pwallx)
                                {
                                    texdelta = -(pwallposi<<10);
                                    xintercept=xtile<<TILESHIFT;
                                    ytile = (short) (yintercept >> TILESHIFT);
                                    tilehit=pwalltile;
                                    HitVertWall();
                                }
                                else
                                {
                                    if(pwalldir==di_south && (int32_t)((word)yintercept)+ystep>(pwallposi<<10)
                                            || pwalldir==di_north && (int32_t)((word)yintercept)+ystep<(pwallposi<<10))
                                        goto passvert;

                                    if(pwalldir==di_south)
                                        yintercept=(yintercept&0xffff0000)-((64-pwallpos)<<10);
                                    else
                                        yintercept=(yintercept&0xffff0000)+((64-pwallpos)<<10);
                                    xintercept=xintercept-((xstep*pwallpos)>>6);
                                    xtile = (short) (xintercept >> TILESHIFT);
                                    tilehit=pwalltile;
                                    HitHorizWall();
                                }
                            }
                        }
                    }
                    else
                    {
                        xintercept=xtile<<TILESHIFT;
                        ytile = (short) (yintercept >> TILESHIFT);
                        HitVertWall();
                    }
                }
                break;
            }
passvert:
            *((byte *)spotvis+xspot)=1;
            xtile+=xtilestep;
            yintercept+=ystep;
            xspot=(word)((xtile<<mapshift)+((uint32_t)yintercept>>16));
        }
        while(1);
        continue;

        do
        {
            if(xtilestep==-1 && (xintercept>>16)<=xtile) goto vertentry;
            if(xtilestep==1 && (xintercept>>16)>=xtile) goto vertentry;
horizentry:
            if((uint32_t)xintercept>mapwidth*65536-1 || (word)ytile>=mapheight)
            {
                if(ytile<0) yintercept=0, ytile=0;
                else if(ytile>=mapheight) yintercept=mapheight<<TILESHIFT, ytile=mapheight-1;
                else ytile=(short) (yintercept >> TILESHIFT);
                if(xintercept<0) xintercept=0, xtile=0;
                else if(xintercept>=(mapwidth<<TILESHIFT)) xintercept=mapwidth<<TILESHIFT, xtile=mapwidth-1;
                xspot=0xffff;
                tilehit=0;
                HitVertBorder();
                break;
            }
            if(yspot>=maparea) break;
            tilehit=((uint16_t *)tilemap)[yspot];
            if(tilehit)
            {
                if(tilehit&BIT_DOOR)
                {
                    int32_t xintbuf=xintercept+(xstep>>1);
                    if((xintbuf>>16)!=(xintercept>>16))
                        goto passhoriz;
                    if((word)xintbuf<doorposition[tilehit&LAST_DOORNUM])
                        goto passhoriz;
                    xintercept=xintbuf;
                    yintercept=(ytile<<TILESHIFT)+0x8000;
                    xtile = (short) (xintercept >> TILESHIFT);
                    HitHorizDoor();
                }
                else
                {
                    if(tilehit==BIT_WALL)
                    {
                        if(pwalldir==di_north || pwalldir==di_south)
                        {
                            int32_t xintbuf;
                            int pwallposnorm;
                            int pwallposinv;
                            if(pwalldir==di_north)
                            {
                                pwallposnorm = 64-pwallpos;
                                pwallposinv = pwallpos;
                            }
                            else
                            {
                                pwallposnorm = pwallpos;
                                pwallposinv = 64-pwallpos;
                            }
                            if(pwalldir == di_south && ytile==pwally && ((uint32_t)xintercept>>16)==pwallx
                                || pwalldir == di_north && !(ytile==pwally && ((uint32_t)xintercept>>16)==pwallx))
                            {
                                xintbuf=xintercept+((xstep*pwallposnorm)>>6);
                                if((xintbuf>>16)!=(xintercept>>16))
                                    goto passhoriz;

                                yintercept=(ytile<<TILESHIFT)+TILEGLOBAL-(pwallposinv<<10);
                                xintercept=xintbuf;
                                xtile = (short) (xintercept >> TILESHIFT);
                                tilehit=pwalltile;
                                HitHorizWall();
                            }
                            else
                            {
                                xintbuf=xintercept+((xstep*pwallposinv)>>6);
                                if((xintbuf>>16)!=(xintercept>>16))
                                    goto passhoriz;

                                yintercept=(ytile<<TILESHIFT)-(pwallposinv<<10);
                                xintercept=xintbuf;
                                xtile = (short) (xintercept >> TILESHIFT);
                                tilehit=pwalltile;
                                HitHorizWall();
                            }
                        }
                        else
                        {
                            int pwallposi = pwallpos;
                            if(pwalldir==di_west) pwallposi = 64-pwallpos;
                            if(pwalldir==di_east && (word)xintercept<(pwallposi<<10)
                                    || pwalldir==di_west && (word)xintercept>(pwallposi<<10))
                            {
                                if(((uint32_t)xintercept>>16)==pwallx && ytile==pwally)
                                {
                                    if(pwalldir==di_east && (int32_t)((word)xintercept)+xstep<(pwallposi<<10)
                                            || pwalldir==di_west && (int32_t)((word)xintercept)+xstep>(pwallposi<<10))
                                        goto passhoriz;

                                    if(pwalldir==di_east)
                                        xintercept=(xintercept&0xffff0000)+(pwallposi<<10);
                                    else
                                        xintercept=(xintercept&0xffff0000)-TILEGLOBAL+(pwallposi<<10);
                                    yintercept=yintercept-((ystep*(64-pwallpos))>>6);
                                    ytile = (short) (yintercept >> TILESHIFT);
                                    tilehit=pwalltile;
                                    HitVertWall();
                                }
                                else
                                {
                                    texdelta = -(pwallposi<<10);
                                    yintercept=ytile<<TILESHIFT;
                                    xtile = (short) (xintercept >> TILESHIFT);
                                    tilehit=pwalltile;
                                    HitHorizWall();
                                }
                            }
                            else
                            {
                                if(((uint32_t)xintercept>>16)==pwallx && ytile==pwally)
                                {
                                    texdelta = -(pwallposi<<10);
                                    yintercept=ytile<<TILESHIFT;
                                    xtile = (short) (xintercept >> TILESHIFT);
                                    tilehit=pwalltile;
                                    HitHorizWall();
                                }
                                else
                                {
                                    if(pwalldir==di_east && (int32_t)((word)xintercept)+xstep>(pwallposi<<10)
                                            || pwalldir==di_west && (int32_t)((word)xintercept)+xstep<(pwallposi<<10))
                                        goto passhoriz;

                                    if(pwalldir==di_east)
                                        xintercept=(xintercept&0xffff0000)-((64-pwallpos)<<10);
                                    else
                                        xintercept=(xintercept&0xffff0000)+((64-pwallpos)<<10);
                                    yintercept=yintercept-((ystep*pwallpos)>>6);
                                    ytile = (short) (yintercept >> TILESHIFT);
                                    tilehit=pwalltile;
                                    HitVertWall();
                                }
                            }
                        }
                    }
                    else
                    {
                        yintercept=ytile<<TILESHIFT;
                        xtile = (short) (xintercept >> TILESHIFT);
                        HitHorizWall();
                    }
                }
                break;
            }
passhoriz:
            *((byte *)spotvis+yspot)=1;
            ytile+=ytilestep;
            xintercept+=xstep;
            yspot=(word)((((uint32_t)xintercept>>16)<<mapshift)+ytile);
        }
        while(1);
    }
}

/*
====================
=
= WallRefresh
=
====================
*/

void WallRefresh (void)
{
    xpartialdown = viewx&(TILEGLOBAL-1);
    xpartialup = TILEGLOBAL-xpartialdown;
    ypartialdown = viewy&(TILEGLOBAL-1);
    ypartialup = TILEGLOBAL-ypartialdown;

    min_wallheight = viewheight;
    lastside = -1;                  // the first pixel is on a new wall
    AsmRefresh ();
    ScalePost ();                   // no more optimization on last post
}

void CalcViewVariables()
{
    viewangle = player->angle;
    midangle = viewangle*(FINEANGLES/ANGLES);
    viewsin = sintable[viewangle];
    viewcos = costable[viewangle];
    viewx = player->x - FixedMul(focallength,viewcos);
    viewy = player->y + FixedMul(focallength,viewsin);

    focaltx = (short)(viewx>>TILESHIFT);
    focalty = (short)(viewy>>TILESHIFT);

    viewtx = (short)(player->x >> TILESHIFT);
    viewty = (short)(player->y >> TILESHIFT);
}

//==========================================================================
void BonusFlashes()
{
///////
// godmode on/warning flash
///////
    if (gamestate.godmodecount > 4*70+55 || 
        (gamestate.godmodecount<=4*70 && gamestate.godmodecount>=3*70+55) || 
        (gamestate.godmodecount<=3*70 && gamestate.godmodecount>=2*70+55) || 
        (gamestate.godmodecount<=2*70 && gamestate.godmodecount>=1*70+55) || 
        (gamestate.godmodecount<=1*70 && gamestate.godmodecount>=55))
        GodInvertFlash();
///////
// invulnerability on/warning flash
///////
    if (gamestate.invisibilitycount > 4*70+55 || 
        (gamestate.invisibilitycount<=4*70 && gamestate.invisibilitycount>=3*70+55) || 
        (gamestate.invisibilitycount<=3*70 && gamestate.invisibilitycount>=2*70+55) || 
        (gamestate.invisibilitycount<=2*70 && gamestate.invisibilitycount>=1*70+55) || 
        (gamestate.invisibilitycount<=1*70 && gamestate.invisibilitycount>=55))
        gamestate.startinvistrigger=true;
    else
        gamestate.startinvistrigger=false;
///////
// berserk red
///////
    if (gamestate.berserkcount > 0 && gamestate.berserkcount <= 8*70)
        StartDamageFlash(3/2);
    else if (gamestate.berserkcount > 8*70  && gamestate.berserkcount <= 20*70)
        StartDamageFlash(-1);
///////
// biosuit on/warning flash
///////
    if (gamestate.biosuitcount > 4*70+55 || 
        (gamestate.biosuitcount<=4*70 && gamestate.biosuitcount>=3*70+55) || 
        (gamestate.biosuitcount<=3*70 && gamestate.biosuitcount>=2*70+55) || 
        (gamestate.biosuitcount<=2*70 && gamestate.biosuitcount>=1*70+55) || 
        (gamestate.biosuitcount<=1*70 && gamestate.biosuitcount>=55))
        StartBiosuitFlash ();
}
/*
========================
=
= ThreeDRefresh
=
========================
*/
void DrawLevelName();
void    ThreeDRefresh (void)
{
//
// clear out the traced array
//
    memset(spotvis,0,maparea);
    spotvis[player->tilex][player->tiley] = 1;       // Detect all sprites over player fix

    vbuf = VL_LockSurface(screenBuffer);
    if(vbuf == NULL) return;

    vbuf += screenofs;
    vbufPitch = bufferPitch;

    CalcViewVariables();

//
// follow the walls from there to the right, drawing as we go
//
    VGAClearScreen ();
    WallRefresh ();
    DrawParallax(vbuf, vbufPitch);
    DrawFloorAndCeiling(vbuf, vbufPitch, min_wallheight);

//
// draw all the scaled images
//
    DrawScaleds();                  // draw scaled stuff
    DrawPlayerWeapon ();    // draw player's hands
    
    if(viewsize == 21 && gamestate.weapon != -1)
        ShowActStatus();
    else if (viewsize < 21)
        DrawPlayScreen();
    
    if (messagetime > 0)
        DrawMessage (); 

    if (Keyboard[sc_Tab]) {
        DrawLevelName();
    }

#ifdef DEBUGGING
    fpscounter = Keyboard[sc_O];
#endif

    
    BonusFlashes();

    VL_UnlockSurface(screenBuffer);
    vbuf = NULL;
    frameon++;

//
// show screen and time last cycle
//

    if (fizzlein)
    {
        FizzleFade(screenBuffer, 0, 0, screenWidth, screenHeight, 20, false);
        fizzlein = false;

        lasttimecount = GetTimeCount();          // don't make a big tic count
    }
    else
    {
#ifdef DEBUGGING
        if (fpscounter)
        {
            fontnumber = 0;
            SETFONTCOLOR(0xb5,0);
            PrintX=274; PrintY=1;
            VWB_Bar(270,0,50,10,0);
            US_PrintUnsigned(fps);
            LSWrite(292,2,"fps");
        }
#endif
        SDL_BlitSurface(screenBuffer, NULL, screen, NULL);
        SDL_Flip(screen);
    }

#ifdef DEBUGGING
    if (fpscounter)
    {
        fps_frames++;
        fps_time+=tics;

        if(fps_time>35)
        {
            fps_time-=35;
            fps=fps_frames<<1;
            fps_frames=0;
        }
    }
#endif
}

/*
========================
=
= GetMessage
=
= gets ingame messages
=
========================
*/

void GetMessage (char *lastmessage)
{
   messagetime = 210; // time for message to display

   strcpy(gamestate.message, lastmessage);
}   

/*
========================
=
= DrawMessage
=
= displays ingame messages
=
========================
*/

void DrawMessage (void)
{
    messagetime-=tics;
   
    CacheLump (MESSAGE_LUMP_START,MESSAGE_LUMP_END);
    LSWrite (2,2,gamestate.message);
    UnCacheLump (MESSAGE_LUMP_START,MESSAGE_LUMP_END);
} 

void DrawLevelName ()
{
    CacheLump (MESSAGE_LUMP_START,MESSAGE_LUMP_END);
    LSWrite (2, 160, levelNameList[gamestate.episode * 10 + gamestate.mapon]);
    UnCacheLump (MESSAGE_LUMP_START,MESSAGE_LUMP_END);
}